@@ -1015,6 +1015,15 @@ public class ContextManager {
 	Context ctx = req.getContext();
 	if(ctx==null) ctx=getContext("");
 
+	if(ctx == null){
+		// The request didn't map into any context so send a 404 error.
+		try{
+			handleContextNotFound(req, res);
+		}catch(IOException e){
+		}
+		return;
+	}
+
 	// don't log normal cases ( redirect and need_auth ), they are not
 	// error
 	// XXX this log was intended to debug the status code generation.
@@ -1073,8 +1082,12 @@ public class ContextManager {
 	    the client. AFAIK the trace is the _best_ debugger.
 	*/
 	if (t instanceof UnavailableException) {
-	    int unavailableTime = ((UnavailableException)t).getUnavailableSeconds();
-	    if( unavailableTime > 0 ) {
+	    int unavailableTime = -1;
+	    if ( !((UnavailableException)t).isPermanent() ) {
+		unavailableTime = ((UnavailableException)t).getUnavailableSeconds();
+		// if unavailable time not known, use 1 second
+		if ( unavailableTime <= 0 )
+		    unavailableTime = 1;
 		res.setHeader("Retry-After", Integer.toString(unavailableTime));
 	    }
 	    String msg=t.getMessage();
@@ -1159,6 +1172,7 @@ public class ContextManager {
 	RequestImpl req1=new RequestImpl();
 	ResponseImpl res1=new ResponseImpl();
 	initRequest( req1, res1 );
+	req1.setServerName( ctx.getHost() );
 
 	req1.setRequestURI( ctx.getPath() + path );
 	processRequest( req1 );
@@ -1182,6 +1196,63 @@ public class ContextManager {
 	return false;
     }
 
+	 /**
+	  *  Called if the request does not map into any context.  This code
+	  * was lifted from DefaultCMSetter.NotFoundHandler becuase without
+	  * a valid context we really can't execute that servlet.
+	  */
+	 private void  handleContextNotFound(Request req, Response res) throws IOException
+	 {
+		 StringManager sm=StringManager.getManager("org.apache.tomcat.resources");
+		 res.setContentType("text/html");	// ISO-8859-1 default
+
+		 String requestURI = (String)req.
+			  getAttribute("javax.servlet.include.request_uri");
+
+		 if (requestURI == null || res.isIncluded()) {
+			  requestURI = req.getRequestURI();
+		 }
+
+		 StringBuffer buf = new StringBuffer();
+		 buf.append("<head><title>")
+			  .append(sm.getString("defaulterrorpage.notfound404"))
+			  .append("</title></head>\r\n");
+		 buf.append("<body><h1>")
+			  .append(sm.getString("defaulterrorpage.notfound404"))
+			  .append("</h1>\r\n<b>");
+		 buf.append(sm.getString("defaulterrorpage.originalrequest"))
+			  .append("</b> ")
+			  .append(RequestUtil.filter(requestURI));
+
+		 if (getShowDebugInfo()) {
+			  if (res.isIncluded()) {
+			 requestURI = (String)req.
+				  getAttribute("javax.servlet.include.request_uri");
+			  }
+			  if (requestURI != null) {
+			 buf.append("<br><br>\r\n<b>")
+				  .append(sm.getString("defaulterrorpage.notfoundrequest"))
+				  .append("</b> ")
+				  .append(RequestUtil.filter(requestURI));
+			  }
+		 }
+
+		 buf.append("</body>\r\n");
+
+		 String body = buf.toString();
+
+		 res.setContentLength(body.length());
+
+		 if( res.isUsingStream() ) {
+			  ServletOutputStream out = res.getOutputStream();
+			  out.print(body);
+			  out.flush();
+		 } else {
+			  PrintWriter out = res.getWriter();
+			  out.print(body);
+			  out.flush();
+		 }
+	  }
 
     // -------------------- Support for notes --------------------
 
